from django import forms
from django.conf import settings

# resource actionform provides deault functionality for this resource:
# from example_app.actions.mixins import TodoActionFormMixin
from actions.forms.mixins import ActionFormMixin
from actions.fields import CommaSeparatedCharField
from example_app.models import Todo

class MarkDoneForm(ActionFormMixin, forms.Form):
    """
    action_id: mark_done
    """
    action_id = 'todo:mark_done'
        # fields go here:
    # ...
    def get_object(self, pk):
        return Todo.objects.get(pk=pk)

    def get_stream_message(self, todo):
        return 'mark_done todo {}'.format(
            todo.title
        )

    def save(self, request, pk=None):

        todo = self.get_object(pk)
        # self.verify_has_permissions(request.user, [todo])

        ## Implement your action here
        todo.status = 'D'
        todo.save()

        # self.perform_post_process(request, todo)
        # extra info to send back to the client, e.g.: count, batch_id etc
        meta = {}
        return (meta, todo)


from django.urls import reverse

def get_instance_action_url(resource, pk, action_id, context=None):
    """
from actions.tests.testutils import get_instance_action_url
get_instance_action_url(resource, pk, 'action_id')
    """
    url = "{}/{}/actions/{}/".format(resource, pk, action_id)
    if context is not None:
        return "/{}{}".format(context, url)
    return url

def get_bulk_action_url(resource, action_id, context=None):
    url = "/{}/actions/{}/".format(resource, action_id)
    if context is not None:
        return "/{}{}".format(context, url)
    return url

def assert_in_registry(client, resource, bulk_actions=[], instance_actions=[]):
    '''
    validate the the required actions are available in the registry
    '''
    url = '/{}/actions/list/'.format(resource)
    res = client.get(url)
    data = res.json()
    bulk = [action.get('id') for action in data.get('bulk_actions')]
    instance = [action.get('id') for action in data.get('instance_actions')]
    for bulk_action in bulk_actions:
        assert bulk_action in bulk, \
            'Expected to find {} registered in: {}'.format(bulk_action, bulk)

    for instance_action in instance_actions:
        assert instance_action in instance, \
            'Expected to find {} registered in: {}'.format(instance_action, instance)




class ActionFormMixin:

    def post_process(self, channel, request, context_object, serializer, custom_message):
        """
        stream_message = self.get_stream_message(request, view, appointment)
        self.post_process(
            appointment.practitioner.channel,
            request,
            appointment,
            AppointmentListSerializer,
            stream_message
        )
        """
        # push to cache
        # todo
        if getattr(self, 'action_id', None) is None:
            raise Exception('Please define `action_id` on you Form. action_id should match the action_id from the action registry')

        context = serializer(context_object).data
        actor_id = request.user.id

        ##
        # we should write result to db or something?


from django.test import TestCase
from example_app.models import Todo
from datetime import date, datetime
from .testutils import (
    assert_in_registry,
    get_instance_action_url,
    get_bulk_action_url
)


class ReadActionsTestCase(TestCase):

    def setUp(self):
        pass

    def test_list_registry(self):
        bulk_actions = [
        ]
        instance_actions = [
            'mark_done'
        ]
        assert_in_registry(self.client, 'todos', bulk_actions, instance_actions)

class MutateActionsTestCase(TestCase):

    def setUp(self):
        self.todo = Todo.objects.create(title='foo', due=date.today(), due_time=datetime.now())
        self.todo2 = Todo.objects.create(title='bar', due=date.today(), due_time=datetime.now())

    # def test_perform_instance_action(self):
    #     url = get_instance_action_url('todos', self.todo.id, 'mark_done')
    #     data = {'status': 'C'}
    #     res = self.client.post(url, data)
    #     self.assertEqual(res.status_code, 200)

    # def test_perform_legacy_instance_action(self):
    #     url = get_instance_action_url('todos', self.todo.id, 'mark_done', context='legacy')
    #     data = {'status': 'C'}
    #     res = self.client.post(url, data)
    #     self.assertEqual(res.status_code, 200)

    # def test_perform_instance_action(self):
    #     url = get_bulk_action_url('todos', 'mark_donw')
    #     data = {
    #         'todos': '{},{}'.format(self.todo.id, self.todo2.id),
    #         'status': 'C'
    #     }
    #     res = self.client.post(url, data)

    # def test_legacy_perform_instance_action(self):
    #     url = get_bulk_action_url('todos', 'mark_donw', context='legacy')
    #     data = {
    #         'todos': '{},{}'.format(self.todo.id, self.todo2.id),
    #         'status': 'C'
    #     }
    #     res = self.client.post(url, data)



# Generated by Django 3.0.3 on 2020-02-25 23:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('example_app', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='todo',
            name='due',
            field=models.DateField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='todo',
            name='due_time',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='todo',
            name='status',
            field=models.CharField(choices=[('N', 'Todo'), ('P', 'In Progress'), ('D', 'Done')], default='N', max_length=30),
        ),
    ]



from django.http import Http404
from django.utils.module_loading import import_string
from rest_framework import response
from django.conf import settings

from .helpers import (
    record_action_start,
    record_action_end
)

import json, time

class BaseActionMixin():
    '''
    ## Perform Action (POST)
    Bulk actions:
    POST /resource/action/:action_id/
    Instance actions:
    POST /resource/:pk/action/:action_id/

    ## Docs: (GET)
    Bulk actions:
    GET /resource/action/:action_id/
    Instance actions:
    GET /resource/:pk/action/:action_id/
    '''
    def __get_form(self, definition, request, pk = None, instance = None):
        form_path = definition.get('form')
        form = import_string(form_path)
        has_custom_getter = getattr(form, 'get_form', None) is not None

        if has_custom_getter:
            return form.get_form(
                self,
                request,
                pk,
                instance = instance
            )
        else:
            return form(request.data)

    def __docs_result(self, request, pk=None, action_id=None):
        is_bulk_action = (pk is None)
        if action_id is None:
            data = self.action_registry
        else:
            data = self.__get_action_definition(action_id, is_bulk_action)
            if data is None:
                available_actions = [key for key, value in self.action_registry.get('actions',{}).items()]
                raise Exception('No action registered with name: {}. Available actions are: '.format(action_id, available_actions))

            if pk is not None:
                url = "{}{}/actions/{}/".format(
                    self.action_registry.get("basepath"),
                    pk,
                    action_id
                )
            else:
                url = "{}/actions/{}/".format(
                    self.action_registry.get("basepath"),
                    action_id
                )

            form = import_string(data.get('form'))()
            fields = []
            for key, value in form.fields.items():
                fields.append({
                    "name": key,
                    "required": value.required,
                    "description": value.help_text,
                    "type": value.__class__.__name__
                })

            serializer_path = data.get('serializer', None)
            if serializer_path is None:
                serializer = self.get_serializer_class()
            else:
                serializer = import_string(serializer_path)

            data.update({
                "url": url,
                "fields": fields,
                "example_response": [serializer().data]
            })
        return response.Response(data)

    def __get_action_definition(self, action_id, is_bulk_action):
        data = self.action_registry.get('actions', {}).get(action_id)
        if is_bulk_action:
            assert data.get('type') == 'bulk', 'No bulk action with that action_id. Try instance action'
        else:
            assert data.get('type') == 'instance', 'No instance action with that action_id. Try bulk action'
        return data

    def _docs(self):
        reg = json.loads(json.dumps(self.action_registry))
        all_actions = reg.pop('actions')
        reg.update({
            "bulk_actions": [],
            "instance_actions": []
        })
        for action_id, action_defn in all_actions.items():
            action_defn['id'] = action_id
            if (action_defn.get('type') == 'bulk'):
                action_defn['url'] = '{}/{}/'.format(reg.get('basepath'), action_id)
                reg['bulk_actions'].append(action_defn)
            else:
                action_defn['url'] = '{}:pk/{}/'.format(reg.get('basepath'), action_id)
                reg['instance_actions'].append(action_defn)

        return response.Response(reg)

    def _action(self, request, pk=None, action_id=None):
        is_bulk_action = (pk is None)
        action_config = self.__get_action_definition(action_id, is_bulk_action)
        if request.method == 'GET':
            return self.__docs_result(request, pk, action_id)
        if action_config is None:
            raise Http404('No action configuration exists for this action_id')

        # handle POST:
        form = None

        if is_bulk_action:
            form = self.__get_form(action_config, request, pk=None, instance=None)
        else:
            instance = self.get_object()
            form = self.__get_form(action_config, request, pk=pk, instance=instance)

        start, actionevent = record_action_start(request, action_id, pk)
        action_response = None
        if form.is_valid():
            meta, result = form.save(request, pk)

            # serializer_class = action_config.get('serializer')
            serializer_path = action_config.get('serializer', None)
            if serializer_path is None:
                serializer = self.get_serializer_class()
            # they explicitly do not want a serializer
            elif serializer_path is False:
                serializer = False
            else:
                serializer = import_string(serializer_path)

            has_multiple_results = action_config.get('multiple_results', is_bulk_action)
            if serializer:
                result = serializer(
                    result,
                    many=has_multiple_results,
                    context={'request': request}
                ).data

            if has_multiple_results:
                http_result = {
                    "meta": meta,
                    "results": result
                }
                action_response = response.Response(http_result)
            else:
                result.update(meta)
                action_response = response.Response(result)
        else:
            action_response = response.Response(form.errors, status=400)

        record_action_end(start, actionevent, action_response)
        return action_response



from django.core.management.base import BaseCommand
import os
from .helpers import template_to_out

class Command(BaseCommand):

    help = 'Configure defaul actions setup inside a module'

    def handle(self, *args, **options):

        # create directory structure
        base_path = '/code/actions/actions/'

        context = {
            "resource": "ActionEvent",
            "code_path": base_path,
            "resource_path": '/dj-actions/actionevents/',
            "appname": "actions"
        }

        try:
            os.makedirs(base_path)
        except FileExistsError:
            print ('actions folder already exists')
        except OSError as e:
            raise CommandError(e)

        # create registry:
        template_to_out(
            'djactions/action_registry.py',
            context,
            '{}action_registry.py'.format(base_path)
        )

        # print instructions
        instructions = template_to_out('djactions/instructions.txt', context)
        print(instructions)



# Generated by Django 2.1.7 on 2019-03-08 16:03

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Todo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner', models.CharField(max_length=30)),
                ('title', models.CharField(max_length=30)),
                ('status', models.CharField(choices=[('N', 'Todo'), ('P', 'In Progress'), ('D', 'Done')], max_length=30)),
                ('size', models.PositiveIntegerField(default=1)),
                ('due', models.DateField()),
                ('due_time', models.DateTimeField()),
            ],
        ),
    ]



from actions.util import (
    STATUS,
    ACTION_TYPE,
    # get_action_docs,
    enrich_actions
)

action_list = []
## -- START ACTION REGISTRY -- ##
action_list.append(('mark_done', ACTION_TYPE.INSTANCE.value, STATUS.ALPHA.value))
## -- END ACTION REGISTRY -- ##

base = '/code/example_app/'
ACTIONS = enrich_actions(
    base,
    { "resource": "todo", "basepath": "/todos/", "actions": {}, },
    action_list
)

print('example_app.actions')
print([action_id for action_id, action_config in ACTIONS.get('actions').items()])
print('------------------------')




from rest_framework import routers, viewsets, serializers
from actions.mixins import ActionMixin
from actions.actions.action_registry import ACTIONS
from actions.models import ActionEvent


class ActionEventSerializer(serializers.Serializer):
    class Meta:
        model = ActionEvent
        fields = '__all__'


class ActionEventViewSet(ActionMixin, viewsets.ModelViewSet):
    queryset = ActionEvent.objects.all()
    serializer_class = ActionEventSerializer
    action_registry = ACTIONS


router = routers.DefaultRouter()
router.register(r'actionevents', ActionEventViewSet)



from .base import BaseActionMixin
from rest_framework import decorators, response
import json

class ActionMixin(BaseActionMixin):
    '''
    Bulk actions:
    POST /resource/action/:action_id/
    Instance actions:
    POST /resource/:pk/action/:action_id/

    ## View docs:

    Bulk actions:
    GET /resource/action/:action_id/
    Instance actions:
    GET /resource/:pk/action/:action_id/
    '''

    @decorators.action(
        detail=False,
        methods=['get'],
        url_path='actions/list'
    )
    def action_docs(self, request, action_id=None):
        reg = json.loads(json.dumps(self.action_registry))

        all_actions = reg.pop('actions')
        reg.update({
            "bulk_actions": [],
            "instance_actions": []
        })
        for action_id, action_defn in all_actions.items():
            action_defn['id'] = action_id
            if (action_defn.get('type') == 'bulk'):
                action_defn['url'] = '{}/{}/'.format(reg.get('basepath'), action_id)
                reg['bulk_actions'].append(action_defn)
            else:
                action_defn['url'] = '{}:pk/{}/'.format(reg.get('basepath'), action_id)
                reg['instance_actions'].append(action_defn)

        return response.Response(reg)

    @decorators.action(
        detail=False,
        methods=['get', 'post'],
        url_path='actions/(?P<action_id>[^/.]+)'
    )
    def bulk_actions(self, request, action_id=None):
        return self._action(request, action_id=action_id, pk=None)

    @decorators.action(
        detail=True,
        methods=['get','post'],
        url_path='actions/(?P<action_id>[^/.]+)'
    )
    def actions(self, request, pk=None, action_id=None):
        return self._action(request, pk, action_id)



from django.test import TestCase
from django.contrib.auth import get_user_model
from actions.tests.testutils import (
    get_instance_action_url,
    get_bulk_action_url
)
from actions.models import ActionEvent
from example_app.models import Todo
from django.contrib.auth import get_user_model

class TestMarkDoneApiTestCase(TestCase):

    def perform_action(self, action_id, payload, performed_by):
        url = get_instance_action_url(
            self.resource,
            self.todo.id,
            action_id,
            context = self.context
        )
        return self.client.post(
            url,
            data = payload,
            context_type='application/json'
        )

    def setUp(self):
        self.context = ''
        self.resource = 'todos'
        self.user = get_user_model().objects.create(username='joe', password='soap')
        self.todo = Todo.objects.create(owner=self.user, title='example todo')

    def test_mark_done(self):
        payload = {}
        res = self.perform_action('mark_done', payload, performed_by=self.user)
        evts = ActionEvent.objects.all()
        import ipdb; ipdb.set_trace()


from django.db import models
from django.contrib.postgres.fields import JSONField
from datetime import timedelta, datetime
DEFAULT_TTL_MINUTES = 60*24 # 1 day
from actions.util import ACTION_TYPE

class ActionEvent(models.Model):

    action_id = models.CharField(max_length=100)
    actor_id = models.CharField(max_length=100, null=True, blank=True)
    object_id = models.CharField(max_length=100, null=True, blank=True)
    action_context = models.CharField(max_length=30, null=True, blank=True)
    status_code = models.PositiveIntegerField(default=0)
    request_data = JSONField(null=True, blank=True)
    response_data = JSONField(null=True, blank=True)
    response_content = models.TextField(null=True, blank=True)
    complete = models.BooleanField(default=False)
    duration = models.DecimalField(max_digits=14, decimal_places=10, default=0)

    scheduled_time = models.DateTimeField(null=True, blank=True)  # <- let's consider supporting scheduling actions

    created_date = models.DateTimeField(auto_now_add=True, db_index=True)
    modified_date = models.DateTimeField(auto_now=True, db_index=True)

    @property
    def action_type(self):
        if self.object_id:
            return ACTION_TYPE.INSTANCE.value
        else:
            return ACTION_TYPE.BULK.value

    @staticmethod
    def archive(self):
        '''
        Archive old action events
        '''
        ttl = getattr(settings, 'DJACTION_TTL_MINUTES', DEFAULT_TTL_MINUTES)
        archive_before_date = datetime.now() - timedelta(minutes=ttl)
        ActionEvent.objects.filter(
            complete=True,
            created_date__gte=archive_before_date
        ).delete()



"""
python manage.py test example_app.tests.actionforms.test_mark_done_form
"""

from django.test import TestCase
from example_app.actions.mark_done.form import MarkDoneForm
from example_app.actions.action_registry import ACTIONS
from actions.util import get_action_form

from django.contrib.auth import get_user_model
from django.test.client import RequestFactory
import mock

from example_app.models import Todo

# BulkAddCodesForm
# use this constant rather than hard importing the form so
# that this code can operate based solely on the registry
ACTION_FORM = get_action_form(ACTIONS, 'mark_done')

class TestMarkDoneFormTestCase(TestCase):

    def submit_form(self, user, payload, pk=None, url='/', expect_errors=False):
        rf = RequestFactory()
        req = rf.post(url, payload)
        req.user = user
        form = ACTION_FORM(payload)
        if form.is_valid():
            return form.save(req, pk)
        else:
            if not expect_errors:
                assert False, 'Did not expect to get form errors. Got: {}'.format(form.errors)

    def setUp(self):
        self.user = get_user_model().objects.create(username='joe', password='soap')
        self.todo = Todo.objects.create(owner=self.user, title='test')

    def test_successful_form_submission(self):
        payload = {}
        res = self.submit_form(self.user, payload, pk=self.user.pk)

        import ipdb; ipdb.set_trace()
        # assert ...



from django.conf import settings
import time
from .models import ActionEvent


def record_action_start(request, action_id, object_id):
    with_track = getattr(settings, 'TRACK_ACTIONS', True)
    if not with_track:
        return None
    event = ActionEvent()
    event.request_data = request.data
    event.action_id = action_id
    event.object_id = object_id
    has_user = getattr(request, 'user')
    if has_user:
        event.actor_id = request.user.id
    start = time.perf_counter()
    return (start, event)


def record_action_end(start, event, response):
    stop = time.perf_counter()
    with_track = getattr(settings, 'TRACK_ACTIONS', True)
    if not with_track:
        return None

    duration = (stop - start)
    print('duration: ', duration)
    event.duration = duration
    event.complete = True
    event.response_data = response.data
    event.status_code = response.status_code
    event.save()
    return event



# Generated by Django 3.0.3 on 2020-02-25 23:49

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ActionEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action_id', models.CharField(max_length=100)),
                ('actor_id', models.CharField(blank=True, max_length=100, null=True)),
                ('object_id', models.CharField(blank=True, max_length=100, null=True)),
                ('action_context', models.CharField(blank=True, max_length=30, null=True)),
                ('status_code', models.PositiveIntegerField(default=0)),
                ('request_data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('response_data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('response_content', models.TextField(blank=True, null=True)),
                ('complete', models.BooleanField(default=False)),
                ('duration', models.DecimalField(decimal_places=10, default=0, max_digits=14)),
                ('scheduled_time', models.DateTimeField(blank=True, null=True)),
                ('created_date', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified_date', models.DateTimeField(auto_now=True, db_index=True)),
            ],
        ),
    ]



from django.template.loader import render_to_string

def template_to_out(path_to_template, context, file_path = None):
    template_string = render_to_string(
        path_to_template,
        context=context
    )
    if file_path:
        with open(file_path, 'w') as f:
            f.write(template_string)
    return template_string



from enum import Enum
from django.template import engines
import yaml, json, os

class STATUS(Enum):
    TODO = 'todo'
    ALPHA = 'alpha'
    BETA = 'beta'
    PROD = 'prod'

class ACTION_TYPE(Enum):
    BULK = 'bulk'
    INSTANCE = 'instance'

def get_action_form(registry, action_id):
    from django.utils.module_loading import import_string
    return import_string(registry.get('actions').get(action_id).get('form'))

def get_instance_action_path(context, collection, id, action):
    return '/api/v3/{}/{}/{}/actions/{}/'.format(
        context,
        collection,
        id,
        action
    )

def get_action_docs(base_path, action_id):
    paths_to_check = [
        # legacy format:
        '{}/actiondocs/{}.yaml'.format(base_path, action_id),
        # new format:
        '{}/actions/{}/definition.yaml'.format(base_path, action_id),
    ]
    rendered = None
    for filepath in paths_to_check:
        try:
            file = open(filepath)
            parsed_yml = yaml.load(file, Loader=yaml.FullLoader)
            json_string = json.dumps(parsed_yml)
            django_engine = engines['django']
            template = django_engine.from_string(json_string)
            rendered = template.render(context={})
            file.close()
        except FileNotFoundError:
            pass
    if rendered is None:
        raise Exception('Could not find definition yaml file. Tried: '.format(paths_to_check))
    return json.loads(rendered)

def enrich_actions(base_path, ACTIONS, action_list, environment = {}):
    for action, action_type, action_status in action_list:
        data = get_action_docs(base_path, action)
        data.update(environment)
        data.update({
            "id": action,
            "type": action_type,
            "status": action_status,
        })
        ACTIONS.get('actions').update({action: data})
    return ACTIONS


import setuptools

with open("./README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dj-actions",
    version="0.0.2",
    author="Christo Crampton",
    author_email="info@38.co.za",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/schoolorchestration/libs/dj-actions",
    packages=setuptools.find_packages(
        exclude=('example_project', 'example_app',)
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


"""
for versions of DRF prior to 3.10:
https://www.django-rest-framework.org/community/3.8-announcement/#deprecations
"""
from .base import BaseActionMixin
from rest_framework import decorators, response
import json

class ActionMixin(BaseActionMixin):
    @decorators.list_route(
        methods=['get'],
        url_path='actions/list'
    )
    def action_docs(self, request, action_id=None):
        return self._docs()

    @decorators.list_route(
        methods=['get', 'post'],
        url_path='actions/(?P<action_id>[^/.]+)'
    )
    def bulk_actions(self, request, action_id=None):
        return self._action(request, action_id=action_id, pk=None)

    @decorators.detail_route(
        methods=['get','post'],
        url_path='actions/(?P<action_id>[^/.]+)'
    )
    def actions(self, request, pk=None, action_id=None):
        return self._action(request, pk, action_id)
